# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import deque

import numpy as np

from qililab.qprogram.blocks import Average, Block, ForLoop, Loop, Parallel
from qililab.qprogram.operations import (
    Acquire,
    Play,
    ResetPhase,
    SetFrequency,
    SetGain,
    SetOffset,
    SetPhase,
    Sync,
    Wait,
)
from qililab.qprogram.variable import FloatVariable, IntVariable, Variable
from qililab.waveforms import IQPair, Waveform


class QProgram:
    """A class for building quantum programs.

    This class provides an interface for building quantum programs,
    including defining operations, managing variables, and handling blocks.
    It contains methods for creating, manipulating and controlling
    the execution flow of quantum operations within a program.

    Attributes:
        _program (Block): The main program block.
        _variables (list[Variable]): List of variables used within the program.
        _block_stack (deque[Block]): A stack to manage nested blocks within the program.

    Examples:

        The following example illustrates how to define a Rabi sequence using QProgram.

        .. code-block:: python3

            qp = QProgram()
            amplitude = qp.variable(float)
            drag = DRAG(amplitude=amplitude, duration=40, num_sigmas=4, drag_correction=1.2)
            square_wf = Square(amplitude=1.0, duration=1000)
            zeros_wf = Square(amplitude=0.0, duration=1000)
            with qp.loop(variable=amplitude, values=np.arange(0, 1, 101)):
                qp.play(bus="drive", waveform=drag)
                qp.sync()
                qp.play(bus="readout", waveform=IQPair(I=square_wf, Q=zeros_wf))
                qp.acquire(bus="readout")
    """

    def __init__(self):
        self._program: Block = Block()
        self._variables: list[Variable] = []
        self._block_stack: deque[Block] = deque([self._program])

    def _append_to_block_stack(self, block: Block):
        self._block_stack.append(block)

    def _pop_from_block_stack(self):
        return self._block_stack.pop()

    @property
    def _active_block(self) -> Block:
        return self._block_stack[-1]

    def block(self):
        """Define a generic block for scoping operations.

        Blocks need to open a scope.

        Returns:
            Block: The block.

        Examples:

            >>> with qp.block() as block:
            >>>    # operations that shall be executed in the block
        """
        return QProgram._BlockContext(qprogram=self)

    def parallel(self, loops: list[ForLoop]):
        """Define a block for running multiple loops in parallel.

        Blocks need to open a scope.

        Examples:
            >>> gain = qp.variable(float)
            >>> frequency = qp.variable(float)
            >>> with qp.parallel(loops=[ForLoop(variable=frequency, start=0, stop=100, step=10),
                                        ForLoop(variable=gain, start=0.0, stop=1.0, step=0.1)]):
            >>>    # operations that shall be executed in the block

        Args:
            loops (list[Loop  |  ForLoop]): The loops to run in parallel

        Returns:
            Parallel: The parallel block.
        """
        return QProgram._ParallelContext(qprogram=self, loops=loops)

    def average(self, shots: int):
        """Define an acquire loop block with averaging in real time.

        Blocks need to open a scope.

        Args:
            iterations (int): The number of acquire iterations.

        Returns:
            AcquireLoop: The acquire_loop block.

        Examples:

            >>> with qp.acquire_loop(iterations=1000):
            >>>    # operations that shall be executed in the acquire_loop block
        """
        return QProgram._AverageContext(qprogram=self, iterations=shots)

    def loop(self, variable: Variable, values: np.ndarray):
        """Define a loop block to iterate values over a variable.

        Blocks need to open a scope.

        Args:
            variable (Variable): The variable to be affected from the loop.
            values (np.ndarray): The values to iterate over.

        Returns:
            Loop: The loop block.

        Examples:

            >>> variable = qp.variable(int)
            >>> with qp.loop(variable=variable, values=np.array(range(100))):
            >>>    # operations that shall be executed in the loop block
        """

        return QProgram._LoopContext(qprogram=self, variable=variable, values=values)

    def for_loop(self, variable: Variable, start: int | float, stop: int | float, step: int | float = 1):
        """Define a for_loop block to iterate values over a variable.

        Blocks need to open a scope.

        Args:
            variable (Variable): The variable to be affected from the loop.
            start (int | float): The start value.
            stop (int | float): The stop value.
            step (int | float, optional): The step value. Defaults to 1.

        Returns:
            Loop: The loop block.

        Examples:

            >>> variable = qp.variable(int)
            >>> with qp.for_loop(variable=variable, start=0, stop=100, step=5)):
            >>>    # operations that shall be executed in the for_loop block
        """

        return QProgram._ForLoopContext(qprogram=self, variable=variable, start=start, stop=stop, step=step)

    def play(self, bus: str, waveform: Waveform | IQPair):
        """Play a single waveform or an I/Q pair of waveforms on the bus.

        Args:
            bus (str): Unique identifier of the bus.
            waveform (Waveform | IQPair): A single waveform or an I/Q pair of waveforms
        """
        operation = Play(bus=bus, waveform=waveform)
        self._active_block.append(operation)

    def wait(self, bus: str, time: int):
        """Adds a delay on the bus with a specified time.

        Args:
            bus (str): Unique identifier of the bus.
            time (int): Duration of the delay.
        """
        operation = Wait(bus=bus, time=time)
        self._active_block.append(operation)

    def acquire(self, bus: str, weights: IQPair | None = None):
        """Acquire results.

        Args:
            bus (str): Unique identifier of the bus.
            weights (IQPair | None, optional): Weights used during acquisition. Defaults to None.
        """
        operation = Acquire(bus=bus, weights=weights)
        self._active_block.append(operation)

    def sync(self, buses: list[str] | None = None):
        """Synchronize operations between buses, so the operations following will start at the same time. If no buses are given, then the synchronization will involve all buses present in the QProgram.

        Args:
            buses (list[str] | None, optional): List of unique identifiers of the buses. Defaults to None.
        """
        operation = Sync(buses=buses)
        self._active_block.append(operation)

    def reset_phase(self, bus: str):
        """Reset the absolute phase of the NCO associated with the bus.

        Args:
            bus (str): Unique identifier of the bus.
        """
        operation = ResetPhase(bus=bus)
        self._active_block.append(operation)

    def set_phase(self, bus: str, phase: float):
        """Set the absolute phase of the NCO associated with the bus.

        Args:
            bus (str): Unique identifier of the bus.
            phase (float): The new absolute phase of the NCO.
        """
        operation = SetPhase(bus=bus, phase=phase)
        self._active_block.append(operation)

    def set_frequency(self, bus: str, frequency: float):
        """Set the frequency of the NCO associated with bus.

        Args:
            bus (str): Unique identifier of the bus.
            frequency (float): The new frequency of the NCO.
        """
        operation = SetFrequency(bus=bus, frequency=frequency)
        self._active_block.append(operation)

    def set_gain(self, bus: str, gain_path0: float, gain_path1: float):
        """Set the gain of the AWG associated with bus.

        Args:
            bus (str): Unique identifier of the bus.
            gain_path0 (float): The new gain of the AWG for path0.
            gain_path1 (float): The new gain of the AWG for path1.
        """
        operation = SetGain(bus=bus, gain_path0=gain_path0, gain_path1=gain_path1)
        self._active_block.append(operation)

    def set_offset(self, bus: str, offset_path0: float, offset_path1: float):
        """Set the gain of the AWG associated with bus.

        Args:
            bus (str): Unique identifier of the bus.
            offset_path0 (float): The new offset of the AWG for path0.
            offset_path1 (float): The new offset of the AWG for path1.
        """
        operation = SetOffset(bus=bus, offset_path0=offset_path0, offset_path1=offset_path1)
        self._active_block.append(operation)

    def variable(self, type: type[int | float]):  # pylint: disable=redefined-builtin
        """Declare a variable.

        Args:
            type (int | float): The type of the variable.

        Raises:
            NotImplementedError: If an unsupported type is provided.

        Returns:
            IntVariable | FloatVariable: The variable.
        """

        def _int_variable(value: int = 0) -> IntVariable:
            variable = IntVariable(value)
            self._variables.append(variable)
            return variable

        def _float_variable(value: float = 0.0) -> FloatVariable:
            variable = FloatVariable(value)
            self._variables.append(variable)
            return variable

        if type == int:
            return _int_variable()
        if type == float:
            return _float_variable()
        raise NotImplementedError

    class _BlockContext:
        def __init__(self, qprogram: "QProgram"):
            self.qprogram = qprogram
            self.block: Block = Block()

        def __enter__(self):
            self.qprogram._append_to_block_stack(block=self.block)
            return self.block

        def __exit__(self, exc_type, exc_value, exc_tb):
            block = self.qprogram._pop_from_block_stack()
            self.qprogram._active_block.append(block)

    class _ParallelContext(_BlockContext):  # pylint: disable=too-few-public-methods
        def __init__(self, qprogram: "QProgram", loops: list[ForLoop]):  # pylint: disable=super-init-not-called
            self.qprogram = qprogram
            self.block: Parallel = Parallel(loops=loops)

        def __enter__(self) -> Parallel:
            self.qprogram._append_to_block_stack(block=self.block)
            return self.block

    class _LoopContext(_BlockContext):  # pylint: disable=too-few-public-methods
        def __init__(  # pylint: disable=super-init-not-called
            self, qprogram: "QProgram", variable: Variable, values: np.ndarray
        ):
            self.qprogram = qprogram
            self.block: Loop = Loop(variable=variable, values=values)

        def __enter__(self) -> Loop:
            self.qprogram._append_to_block_stack(block=self.block)
            return self.block

    class _ForLoopContext(_BlockContext):  # pylint: disable=too-few-public-methods
        def __init__(  # pylint: disable=super-init-not-called
            self, qprogram: "QProgram", variable: Variable, start: int | float, stop: int | float, step: int | float
        ):
            self.qprogram = qprogram
            self.block: ForLoop = ForLoop(variable=variable, start=start, stop=stop, step=step)

        def __enter__(self) -> ForLoop:
            self.qprogram._append_to_block_stack(block=self.block)
            return self.block

    class _AverageContext(_BlockContext):  # pylint: disable=too-few-public-methods
        def __init__(self, qprogram: "QProgram", iterations: int):  # pylint: disable=super-init-not-called
            self.qprogram = qprogram
            self.block: Average = Average(shots=iterations)

        def __enter__(self) -> Average:
            self.qprogram._append_to_block_stack(block=self.block)
            return self.block



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

from qcodes.instrument import DelegateParameter
from qcodes.instrument_drivers.rohde_schwarz import RohdeSchwarzSGS100A as QcodesSGS100A

from qililab.drivers import parameters
from qililab.drivers.instruments.instrument_driver_factory import InstrumentDriverFactory
from qililab.drivers.interfaces import LocalOscillator


@InstrumentDriverFactory.register
class RhodeSchwarzSGS100A(QcodesSGS100A, LocalOscillator):
    """Qililab's driver for the SGS100A local oscillator

    QcodesSGS100A: QCoDeS Rohde Schwarz driver for SGS100A
    LocalOscillator: Qililab's local oscillator interface
    """

    def __init__(self, name: str, address: str, **kwargs: Any) -> None:
        super().__init__(name, address, **kwargs)

        # change the name for frequency
        self.add_parameter(
            parameters.lo.frequency,
            label="Delegated parameter for local oscillator frequency",
            source=self.parameters["frequency"],
            parameter_class=DelegateParameter,
        )

    @property
    def params(self):
        """return the parameters of the instrument"""
        return self.parameters

    @property
    def alias(self):
        """return the alias of the instrument, which corresponds to the QCodes name attribute"""
        return self.name



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""PulseShape abstract base class."""
from abc import abstractmethod
from dataclasses import dataclass, field

import numpy as np

from qililab.typings import PulseShapeName
from qililab.typings.factory_element import FactoryElement
from qililab.utils import Factory


@dataclass(frozen=True, eq=True)
class PulseShape(FactoryElement):
    """Pulse shapes describe the shape of the :class:`Pulse`'s envelopes. ``PulseShape`` is their abstract base class.

    Every child of this interface needs to contain an `envelope` and `to/from_dict` methods (for serialization).

    The `envelope` method will create the corresponding array of each shape.

    Derived: :class:`Rectangular`, :class:`Gaussian`, :class:`Drag`, :class:`Cosine` and :class:`SNZ`
    """

    name: PulseShapeName = field(init=False)  #: Name of the pulse shape.

    @abstractmethod
    def envelope(self, duration: int, amplitude: float, resolution: float) -> np.ndarray:
        """Computes the amplitudes of the pulse shape envelope.

        Args:
            duration (int): Duration of the pulse (ns).
            amplitude (float): Maximum amplitude of the pulse.
            resolution (float): Resolution of the pulse envelope.

        Returns:
            ndarray: Amplitude of the envelope for each time step.
        """

    @classmethod
    def from_dict(cls, dictionary: dict) -> "PulseShape":
        """Returns dictionary representation of the pulse shape.

        Args:
            dictionary (dict): Dictionary representation of the PulseShape object.

        Returns:
            PulseShape: PulseShape loaded class, including the name of the pulse shape and attributes.
        """
        shape_class = Factory.get(name=dictionary["name"])
        return shape_class.from_dict(dictionary)

    @abstractmethod
    def to_dict(self) -> dict:
        """Returns dictionary representation of the pulse shape.

        Returns:
            dict: Dictionary representation of the pulse shape including the name and attributes of the pulse shape.
        """



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""BusExecution class."""
from dataclasses import dataclass, field

from qililab.platform import Bus
from qililab.pulse import PulseBusSchedule
from qililab.system_control import ReadoutSystemControl, SystemControl
from qililab.utils import Waveforms


# TODO: Remove class once a Drawer class is implemented
@dataclass
class BusExecution:
    """This class contains the information of a specific bus in the platform together with a list of
    pulse schedules that will be executed on this bus.

    This class is a relic and should be removed once the drawing responsibilities are moved to its own class.

    Args:
        bus (Bus): Bus where the pulse schedules will be executed.
        pulse_bus_schedules (list[PulseBusSchedule]): pulse schedules to execute on the Bus.
    """

    bus: Bus
    pulse_bus_schedules: list[PulseBusSchedule] = field(default_factory=list)

    def add_pulse_bus_schedule(self, pulse_bus_schedule: PulseBusSchedule):
        """Add pulse to the BusPulseSequence given by idx.

        Args:
            pulse (Pulse): Pulse object.
            idx (int): Index of the BusPulseSequence to add the pulse.
        """
        self.pulse_bus_schedules.append(pulse_bus_schedule)

    def acquire_time(self, idx: int = 0) -> int:
        """BusExecution 'acquire_time' property.

        Returns:
            int: Acquire time (in ns).
        """
        num_sequences = len(self.pulse_bus_schedules)
        if idx >= num_sequences:
            raise IndexError(f"Index {idx} is out of bounds for pulse_schedule list of length {num_sequences}")
        readout_schedule = self.pulse_bus_schedules[idx]
        time = readout_schedule.timeline[-1].start_time
        if isinstance(self.system_control, ReadoutSystemControl):
            time += self.system_control.acquisition_delay_time
        return time

    def waveforms(self, modulation: bool = True, resolution: float = 1.0, idx: int = 0) -> Waveforms:
        """Return pulses applied on this bus.

        Args:
            resolution (float): The resolution of the pulses in ns.

        Returns:
            Waveforms: Object containing arrays of the I/Q amplitudes
            of the pulses applied on this bus.
        """
        num_sequences = len(self.pulse_bus_schedules)
        if idx >= num_sequences:
            raise IndexError(f"Index {idx} is out of bounds for pulse_sequences list of length {num_sequences}")
        return self.pulse_bus_schedules[idx].waveforms(modulation=modulation, resolution=resolution)

    @property
    def system_control(self) -> SystemControl:
        """BusExecution 'system_control' property.

        Returns:
            SystemControl: bus.system_control
        """
        return self.bus.system_control

    @property
    def alias(self):
        """BusExecution 'alias' property.

        Returns:
            str: alias of the bus
        """
        return self.bus.alias



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Bus class."""
from dataclasses import InitVar, dataclass

from qililab.chip import Chip, Coil, Coupler, Qubit, Resonator
from qililab.constants import BUS, NODE, RUNCARD
from qililab.instruments import Instruments, ParameterNotFound
from qililab.pulse import PulseBusSchedule, PulseDistortion
from qililab.result import Result
from qililab.settings import Settings
from qililab.system_control import ReadoutSystemControl, SystemControl
from qililab.typings import Parameter
from qililab.utils import Factory


class Bus:
    """Bus class.

    Ideally a bus should contain a qubit control/readout and a signal generator, which are connected
    through a mixer for up- or down-conversion. At the end of the bus there should be a qubit or a resonator object,
    which is connected to one or multiple qubits.

    Args:
        targets (list[Qubit | Resonator | Coupler | Coil]): Port target (or targets in case of multiple resonators).
        settings (BusSettings): Bus settings.
    """

    targets: list[Qubit | Resonator | Coupler | Coil]
    """Port target (or targets in case of multiple resonators)."""

    @dataclass
    class BusSettings(Settings):
        """Bus settings.

        Args:
            alias (str): Alias of the bus.
            system_control (SystemControl): System control used to control and readout the qubits of the bus.
            port (str): Alias of the port where bus is connected.
            distortions (list[PulseDistotion]): List of the distortions to apply to the Bus.
            delay (int): Bus delay
        """

        alias: str
        system_control: SystemControl
        port: str
        platform_instruments: InitVar[Instruments]
        distortions: list[PulseDistortion]
        delay: int

        def __post_init__(self, platform_instruments: Instruments):  # type: ignore # pylint: disable=arguments-differ
            if isinstance(self.system_control, dict):
                system_control_class = Factory.get(name=self.system_control.pop(RUNCARD.NAME))
                self.system_control = system_control_class(
                    settings=self.system_control, platform_instruments=platform_instruments
                )
            super().__post_init__()

            self.distortions = [
                PulseDistortion.from_dict(distortion) for distortion in self.distortions if isinstance(distortion, dict)
            ]

    settings: BusSettings
    """Bus settings. Containing the alias of the bus, the system control used to control and readout its qubits, the alias
    of the port where it's connected, the list of the distortions to apply, and its delay.
    """

    def __init__(self, settings: dict, platform_instruments: Instruments, chip: Chip):
        self.settings = self.BusSettings(**settings, platform_instruments=platform_instruments)  # type: ignore
        self.targets = chip.get_port_nodes(alias=self.port)

    @property
    def alias(self):
        """Alias of the bus.

        Returns:
            str: alias of the bus
        """
        return self.settings.alias

    @property
    def system_control(self):
        """Bus 'system_control' property.

        Returns:
            Resonator: settings.system_control.
        """
        return self.settings.system_control

    @property
    def port(self):
        """Bus 'resonator' property.

        Returns:
            Resonator: settings.resonator.
        """
        return self.settings.port

    @property
    def distortions(self):
        """Bus 'distortions' property.

        Returns:
            list[PulseDistortion]: settings.distortions.
        """
        return self.settings.distortions

    @property
    def delay(self):
        """Bus 'delay' property.

        Returns:
            int: settings.delay.
        """
        return self.settings.delay

    def __str__(self):
        """String representation of a bus. Prints a drawing of the bus elements."""
        return f"Bus {self.alias}:  ----{self.system_control}---" + "".join(
            f"--|{target}|----" for target in self.targets
        )

    def __eq__(self, other: object) -> bool:
        """compare two Bus objects"""
        return str(self) == str(other) if isinstance(other, Bus) else False

    @property
    def target_freqs(self):
        """Bus 'target_freqs' property.

        Returns:
            list[float]: Frequencies of the nodes that have frequencies
        """
        return list(
            filter(None, [target.frequency if hasattr(target, NODE.FREQUENCY) else None for target in self.targets])
        )

    def __iter__(self):
        """Redirect __iter__ magic method."""
        return iter(self.system_control)

    def to_dict(self):
        """Return a dict representation of the Bus class."""
        return {
            RUNCARD.ALIAS: self.alias,
            RUNCARD.SYSTEM_CONTROL: self.system_control.to_dict(),
            BUS.PORT: self.port,
            RUNCARD.DISTORTIONS: [distortion.to_dict() for distortion in self.distortions],
            RUNCARD.DELAY: self.delay,
        }

    def set_parameter(self, parameter: Parameter, value: int | float | str | bool, channel_id: int | None = None):
        """Set a parameter to the bus.

        Args:
            parameter (Parameter): parameter settings of the instrument to update
            value (int | float | str | bool): value to update
            channel_id (int | None, optional): instrument channel to update, if multiple. Defaults to None.
        """
        if parameter == Parameter.DELAY:
            self.settings.delay = int(value)
        else:
            try:
                self.system_control.set_parameter(
                    parameter=parameter, value=value, channel_id=channel_id, port_id=self.port
                )
            except ParameterNotFound as error:
                raise ParameterNotFound(
                    f"No parameter with name {parameter.value} was found in the bus with alias {self.alias}"
                ) from error

    def get_parameter(self, parameter: Parameter, channel_id: int | None = None):
        """Gets a parameter of the bus.

        Args:
            parameter (Parameter): parameter settings of the instrument to update
            value (int | float | str | bool): value to update
            channel_id (int | None, optional): instrument channel to update, if multiple. Defaults to None.
        """
        if parameter == Parameter.DELAY:
            return self.settings.delay
        try:
            return self.system_control.get_parameter(parameter=parameter, channel_id=channel_id, port_id=self.port)
        except ParameterNotFound as error:
            raise ParameterNotFound(
                f"No parameter with name {parameter.value} was found in the bus with alias {self.alias}"
            ) from error

    def compile(
        self, pulse_bus_schedule: PulseBusSchedule, nshots: int, repetition_duration: int, num_bins: int
    ) -> list:
        """Compiles the ``PulseBusSchedule`` into an assembly program.

        Args:
            pulse_bus_schedule (PulseBusSchedule): the list of pulses to be converted into a program
            nshots (int): number of shots / hardware average
            repetition_duration (int): maximum window for the duration of one hardware repetition
            num_bins (int): number of bins
        """
        return self.system_control.compile(pulse_bus_schedule, nshots, repetition_duration, num_bins)

    def upload(self):
        """Uploads any previously compiled program into the instrument."""
        self.system_control.upload(port=self.port)

    def run(self) -> None:
        """Runs any previously uploaded program into the instrument."""
        self.system_control.run(port=self.port)

    def acquire_result(self) -> Result:
        """Read the result from the vector network analyzer instrument

        Returns:
            Result: Acquired result
        """
        if isinstance(self.system_control, ReadoutSystemControl):
            return self.system_control.acquire_result()

        raise AttributeError(
            f"The bus {self.alias} cannot acquire results because it doesn't have a readout system control."
        )



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=protected-access
import math
from collections import deque
from typing import Callable

import numpy as np
import qpysequence as QPy
import qpysequence.program as QPyProgram
import qpysequence.program.instructions as QPyInstructions

from qililab.qprogram.blocks import Average, Block, ForLoop, Loop, Parallel
from qililab.qprogram.operations import (
    Acquire,
    Operation,
    Play,
    ResetPhase,
    SetFrequency,
    SetGain,
    SetOffset,
    SetPhase,
    Sync,
    Wait,
)
from qililab.qprogram.qprogram import QProgram
from qililab.qprogram.variable import Variable
from qililab.waveforms import IQPair, Waveform


class BusInfo:  # pylint: disable=too-many-instance-attributes, too-few-public-methods
    """Class representing the information stored by QBloxCompiler for a bus."""

    def __init__(self):
        # The generated Sequence
        self.qpy_sequence = QPy.Sequence(
            program=QPy.Program(), waveforms=QPy.Waveforms(), acquisitions=QPy.Acquisitions(), weights=QPy.Weights()
        )

        # Dictionaries to hold mappings useful during compilation.
        self.variable_to_register: dict[Variable, QPyProgram.Register] = {}
        self.waveform_to_index: dict[str, int] = {}
        self.weight_to_index: dict[str, int] = {}

        # Create and append the main block to the Sequence's program
        main_block = QPyProgram.Block(name="main")
        self.qpy_sequence._program.append_block(main_block)

        # Stacks to manage block hierarchy during compilation
        self.qpy_block_stack: deque[QPyProgram.Block] = deque([main_block])
        self.qprogram_block_stack: deque[Block] = deque()

        # Counters to help with naming and indexing
        self.next_bin_index = 0
        self.next_acquisition_index = 0
        self.loop_counter = 0
        self.average_counter = 0


class QbloxCompiler:  # pylint: disable=too-few-public-methods
    """A class for compiling QProgram to QBlox hardware."""

    minimum_wait_duration: int = 4

    def __init__(self, integration_length: int = 1000):
        # External settings
        self._integration_length = integration_length

        # Handlers to map each operation to a corresponding handler function
        self._handlers: dict[type, Callable] = {
            Parallel: self._handle_parallel,
            Average: self._handle_average,
            ForLoop: self._handle_for_loop,
            Loop: self._handle_loop,
            SetFrequency: self._handle_set_frequency,
            SetPhase: self._handle_set_phase,
            ResetPhase: self._handle_reset_phase,
            SetGain: self._handle_set_gain,
            SetOffset: self._handle_set_offset,
            Wait: self._handle_wait,
            Sync: self._handle_sync,
            Acquire: self._handle_acquire,
            Play: self._handle_play,
        }

        self._qprogram: QProgram
        self._buses: dict[str, BusInfo]

    def compile(self, qprogram: QProgram) -> dict[str, QPy.Sequence]:
        """Compile QProgram to qpysequence.Sequence

        Args:
            qprogram (QProgram): The QProgram to be compiled

        Returns:
            dict[str, QPy.Sequence]: A dictionary with the buses participating in the QProgram as keys and the corresponding Sequence as values.
        """

        def traverse(block: Block):
            for bus in self._buses:
                self._buses[bus].qprogram_block_stack.append(block)
            for element in block.elements:
                handler = self._handlers.get(type(element))
                if not handler:
                    raise NotImplementedError(f"{element.__class__} is currently not supported in QBlox.")
                appended = handler(element)
                if isinstance(element, Block):
                    traverse(element)
                    if appended:
                        for bus in self._buses:
                            self._buses[bus].qpy_block_stack.pop()
            for bus in self._buses:
                self._buses[bus].qprogram_block_stack.pop()

        self._qprogram = qprogram
        self._buses = self._populate_buses()

        # Recursive traversal to convert QProgram blocks to Sequence
        traverse(self._qprogram._program)

        # Post-processing: Add stop instructions and compile
        for bus in self._buses:
            self._buses[bus].qpy_block_stack[0].append_component(component=QPyInstructions.Stop())
            self._buses[bus].qpy_sequence._program.compile()

        # Return a dictionary with bus names as keys and the compiled Sequence as values.
        return {bus: bus_info.qpy_sequence for bus, bus_info in self._buses.items()}

    def _populate_buses(self):
        """Map each bus in the QProgram to a BusInfo instance.

        Returns:
            A dictionary where the keys are bus names and the values are BusInfo objects.
        """

        def collect_buses(block: Block):
            for element in block.elements:
                if isinstance(element, Block):
                    yield from collect_buses(element)
                if isinstance(element, Operation):
                    bus = getattr(element, "bus", None)
                    if bus:
                        yield bus

        buses = set(collect_buses(self._qprogram._program))
        return {bus: BusInfo() for bus in buses}

    def _append_to_waveforms_of_bus(self, bus: str, waveform_I: Waveform, waveform_Q: Waveform | None):
        """Append waveforms to Sequence's Waveforms of the given bus.

        Args:
            bus (str): Name of the bus.
            waveform_I (Waveform): I waveform.
            waveform_Q (Waveform | None): Q waveform.
        """

        def handle_waveform(waveform: Waveform | None, default_length: int = 0):
            _hash = QbloxCompiler._hash(waveform) if waveform else f"zeros {default_length}"

            if _hash in self._buses[bus].waveform_to_index:
                index = self._buses[bus].waveform_to_index[_hash]
                length = next(
                    len(waveform.data)
                    for waveform in self._buses[bus].qpy_sequence._waveforms._waveforms
                    if waveform.index == index
                )
                return index, length

            envelope = waveform.envelope() if waveform else np.zeros(default_length)
            index = self._buses[bus].qpy_sequence._waveforms.add(envelope)
            self._buses[bus].waveform_to_index[_hash] = index
            return index, len(envelope)

        index_I, length_I = handle_waveform(waveform_I, 0)
        index_Q, length_Q = handle_waveform(waveform_Q, len(waveform_I.envelope()))
        if length_I != length_Q:
            raise NotImplementedError("Waveforms should have equal lengths.")
        return index_I, index_Q, length_I

    def _append_to_weights_of_bus(self, bus: str, weights: IQPair):
        def handle_waveform(waveform: Waveform):
            _hash = QbloxCompiler._hash(waveform)

            if _hash in self._buses[bus].weight_to_index:
                index = self._buses[bus].weight_to_index[_hash]
                length = next(
                    len(weight.data)
                    for weight in self._buses[bus].qpy_sequence._weights._weights  # pylint: disable=protected-access
                    if weight.index == index
                )
                return index, length

            envelope = waveform.envelope()
            length = len(envelope)
            index = self._buses[bus].qpy_sequence._weights.add(envelope)  # pylint: disable=protected-access
            self._buses[bus].weight_to_index[_hash] = index
            return index, length

        index_I, length_I = handle_waveform(weights.I)
        index_Q, length_Q = handle_waveform(weights.Q)
        if length_I != length_Q:
            raise NotImplementedError("Weights should have equal lengths.")
        return index_I, index_Q, length_I

    def _handle_parallel(self, element: Parallel):
        if not element.loops:
            raise NotImplementedError("Parallel block should contain loops.")
        if len({int((loop.stop - loop.start) / loop.step) for loop in element.loops}) != 1:
            raise NotImplementedError("Loops run in parallel should have the same number of iterations.")
        for bus in self._buses:
            iterations = int((element.loops[0].stop - element.loops[0].start) / element.loops[0].step)
            qpy_loop = QPyProgram.Loop(name=f"loop_{self._buses[bus].loop_counter}", begin=iterations)
            for loop in element.loops:
                operation = QbloxCompiler._get_reference_operation_of_loop(loop=loop, starting_block=element)
                if not operation:
                    raise NotImplementedError("Variables referenced in loops should be used in at least one operation.")
                begin, _, step = QbloxCompiler._convert_for_loop_values(loop, operation)
                loop_register = QPyProgram.Register()
                self._buses[bus].qpy_block_stack[-1].append_component(
                    component=QPyInstructions.Move(var=begin, register=loop_register)
                )
                qpy_loop.builtin_components.insert(
                    0,
                    QPyInstructions.Add(loop_register, step, loop_register)
                    if step > 0
                    else QPyInstructions.Sub(loop_register, step, loop_register),
                )
                self._buses[bus].variable_to_register[loop.variable] = loop_register
            qpy_loop.append_component(QPyInstructions.WaitSync(4))
            self._buses[bus].qpy_block_stack[-1].append_component(qpy_loop)
            self._buses[bus].qpy_block_stack.append(qpy_loop)
            self._buses[bus].loop_counter += 1
        return True

    def _handle_average(self, element: Average):
        for bus in self._buses:
            qpy_loop = QPyProgram.Loop(name=f"avg_{self._buses[bus].average_counter}", begin=element.shots)
            qpy_loop.append_component(
                component=QPyInstructions.WaitSync(wait_time=QbloxCompiler.minimum_wait_duration),
                bot_position=len(qpy_loop.components),
            )
            self._buses[bus].qpy_block_stack[-1].append_component(qpy_loop)
            self._buses[bus].qpy_block_stack.append(qpy_loop)
            self._buses[bus].average_counter += 1
        return True

    def _handle_for_loop(self, element: ForLoop):
        operation = QbloxCompiler._get_reference_operation_of_loop(element)
        if not operation:
            raise NotImplementedError("Variables referenced in loops should be used in at least one operation.")
        begin, end, step = QbloxCompiler._convert_for_loop_values(element, operation)
        for bus in self._buses:
            qpy_loop = QPyProgram.Loop(name=f"loop_{self._buses[bus].loop_counter}", begin=begin, end=end, step=step)
            qpy_loop.append_component(
                component=QPyInstructions.WaitSync(wait_time=QbloxCompiler.minimum_wait_duration),
                bot_position=len(qpy_loop.components),
            )
            self._buses[bus].qpy_block_stack[-1].append_component(qpy_loop)
            self._buses[bus].qpy_block_stack.append(qpy_loop)
            self._buses[bus].variable_to_register[element.variable] = qpy_loop.counter_register
            self._buses[bus].loop_counter += 1
        return True

    def _handle_loop(self, element: Loop):
        raise NotImplementedError("Loops with arbitrary numpy arrays are not currently supported for QBlox.")

    def _handle_set_frequency(self, element: SetFrequency):
        convert = QbloxCompiler._convert_value(element)
        frequency = (
            self._buses[element.bus].variable_to_register[element.frequency]
            if isinstance(element.frequency, Variable)
            else convert(element.frequency)
        )
        self._buses[element.bus].qpy_block_stack[-1].append_component(
            component=QPyInstructions.SetFreq(frequency=frequency)
        )

    def _handle_set_phase(self, element: SetPhase):
        convert = QbloxCompiler._convert_value(element)
        phase = (
            self._buses[element.bus].variable_to_register[element.phase]
            if isinstance(element.phase, Variable)
            else convert(element.phase)
        )
        self._buses[element.bus].qpy_block_stack[-1].append_component(component=QPyInstructions.SetPh(phase=phase))

    def _handle_reset_phase(self, element: ResetPhase):
        self._buses[element.bus].qpy_block_stack[-1].append_component(component=QPyInstructions.ResetPh())

    def _handle_set_gain(self, element: SetGain):
        convert = QbloxCompiler._convert_value(element)
        gain_0 = (
            self._buses[element.bus].variable_to_register[element.gain_path0]
            if isinstance(element.gain_path0, Variable)
            else convert(element.gain_path0)
        )
        gain_1 = (
            self._buses[element.bus].variable_to_register[element.gain_path1]
            if isinstance(element.gain_path1, Variable)
            else convert(element.gain_path1)
        )
        self._buses[element.bus].qpy_block_stack[-1].append_component(
            component=QPyInstructions.SetAwgGain(gain_0=gain_0, gain_1=gain_1)
        )

    def _handle_set_offset(self, element: SetOffset):
        convert = QbloxCompiler._convert_value(element)
        offset_0 = (
            self._buses[element.bus].variable_to_register[element.offset_path0]
            if isinstance(element.offset_path0, Variable)
            else convert(element.offset_path0)
        )
        offset_1 = (
            self._buses[element.bus].variable_to_register[element.offset_path1]
            if isinstance(element.offset_path1, Variable)
            else convert(element.offset_path1)
        )
        self._buses[element.bus].qpy_block_stack[-1].append_component(
            component=QPyInstructions.SetAwgOffs(gain_0=offset_0, gain_1=offset_1)
        )

    def _handle_wait(self, element: Wait):
        convert = QbloxCompiler._convert_value(element)
        time = (
            self._buses[element.bus].variable_to_register[element.time]
            if isinstance(element.time, Variable)
            else convert(element.time)
        )
        self._buses[element.bus].qpy_block_stack[-1].append_component(component=QPyInstructions.Wait(wait_time=time))

    def _handle_sync(self, _: Sync):
        # QBlox does not currently support syncing on selective sequences. If it did we would do somethin like:
        # buses = element.buses if element.buses is not None else self._buses.keys()
        for bus in self._buses:
            self._buses[bus].qpy_block_stack[-1].append_component(
                component=QPyInstructions.WaitSync(wait_time=QbloxCompiler.minimum_wait_duration)
            )

    def _handle_acquire(self, element: Acquire):
        loops = [
            (i, loop)
            for i, loop in enumerate(self._buses[element.bus].qpy_block_stack)
            if isinstance(loop, QPyProgram.Loop) and not loop.name.startswith("avg")
        ]
        num_bins = math.prod(loop[1]._iterations for loop in loops)
        self._buses[element.bus].qpy_sequence._acquisitions.add(
            name=f"acquisition_{self._buses[element.bus].next_acquisition_index}",
            num_bins=num_bins,
            index=self._buses[element.bus].next_acquisition_index,
        )

        if element.weights:
            index_I, index_Q, integration_length = self._append_to_weights_of_bus(element.bus, weights=element.weights)

        if num_bins > 1:
            bin_register = QPyProgram.Register()
            block_index_for_move_instruction = loops[0][0] - 1 if loops else -2
            block_index_for_add_instruction = loops[-1][0] if loops else -1
            self._buses[element.bus].qpy_block_stack[block_index_for_move_instruction].append_component(
                component=QPyInstructions.Move(var=self._buses[element.bus].next_bin_index, register=bin_register),
                bot_position=len(self._buses[element.bus].qpy_block_stack[block_index_for_move_instruction].components),
            )
            if element.weights:
                register_I, register_Q = QPyProgram.Register(), QPyProgram.Register()
                self._buses[element.bus].qpy_block_stack[block_index_for_move_instruction].append_component(
                    component=QPyInstructions.Move(var=index_I, register=register_I),
                    bot_position=len(
                        self._buses[element.bus].qpy_block_stack[block_index_for_move_instruction].components
                    ),
                )
                self._buses[element.bus].qpy_block_stack[block_index_for_move_instruction].append_component(
                    component=QPyInstructions.Move(var=index_Q, register=register_Q),
                    bot_position=len(
                        self._buses[element.bus].qpy_block_stack[block_index_for_move_instruction].components
                    ),
                )
                self._buses[element.bus].qpy_block_stack[-1].append_component(
                    component=QPyInstructions.AcquireWeighed(
                        acq_index=self._buses[element.bus].next_acquisition_index,
                        bin_index=bin_register,
                        weight_index_0=register_I,
                        weight_index_1=register_Q,
                        wait_time=integration_length,
                    )
                )
            else:
                self._buses[element.bus].qpy_block_stack[-1].append_component(
                    component=QPyInstructions.Acquire(
                        acq_index=self._buses[element.bus].next_acquisition_index,
                        bin_index=bin_register,
                        wait_time=self._integration_length,
                    )
                )
            self._buses[element.bus].qpy_block_stack[block_index_for_add_instruction].append_component(
                component=QPyInstructions.Add(origin=bin_register, var=1, destination=bin_register)
            )
        elif element.weights:
            self._buses[element.bus].qpy_block_stack[-1].append_component(
                component=QPyInstructions.AcquireWeighed(
                    acq_index=self._buses[element.bus].next_acquisition_index,
                    bin_index=self._buses[element.bus].next_bin_index,
                    weight_index_0=index_I,
                    weight_index_1=index_Q,
                    wait_time=integration_length,
                )
            )
        else:
            self._buses[element.bus].qpy_block_stack[-1].append_component(
                component=QPyInstructions.Acquire(
                    acq_index=self._buses[element.bus].next_acquisition_index,
                    bin_index=self._buses[element.bus].next_bin_index,
                    wait_time=self._integration_length,
                )
            )
        self._buses[element.bus].next_bin_index += num_bins
        self._buses[element.bus].next_acquisition_index += 1

    def _handle_play(self, element: Play):
        waveform_I, waveform_Q = element.get_waveforms()
        variables = element.get_variables()
        if not variables:
            index_I, index_Q, length = self._append_to_waveforms_of_bus(
                bus=element.bus, waveform_I=waveform_I, waveform_Q=waveform_Q
            )
            self._buses[element.bus].qpy_block_stack[-1].append_component(
                component=QPyInstructions.Play(index_I, index_Q, wait_time=length)
            )

    @staticmethod
    def _get_reference_operation_of_loop(loop: ForLoop, starting_block: Block | None = None):
        def collect_operations(block: Block):
            for element in block.elements:
                if isinstance(element, Block):
                    yield from collect_operations(element)
                else:
                    if any(variable is loop.variable for variable in element.get_variables()):
                        yield element

        starting_block = starting_block or loop
        operations = list(collect_operations(starting_block))

        if not operations:
            return None
        if len(set(map(type, operations))) != 1:
            raise NotImplementedError("Variables referenced in a loop cannot be used in different types of operations.")
        if isinstance(operations[0], Play):
            raise NotImplementedError("TODO: Variables referenced in a loop cannot be used in Play operation.")
        return operations[0]

    @staticmethod
    def _convert_for_loop_values(for_loop: ForLoop, operation: Operation):
        convert = QbloxCompiler._convert_value(operation)
        return tuple(convert(value) for value in (for_loop.start, for_loop.stop, for_loop.step))

    @staticmethod
    def _convert_value(operation: Operation):
        conversion_map = {
            SetFrequency: lambda x: int(x * 4),
            SetPhase: lambda x: int(x * 1e9 / 360),
            SetGain: lambda x: int(x * 32_767),
            SetOffset: lambda x: int(x * 32_767),
            Wait: lambda x: max(x, QbloxCompiler.minimum_wait_duration),
        }
        return conversion_map.get(type(operation), lambda x: x)

    @staticmethod
    def _hash(waveform: Waveform):
        hashes = {
            key: (value.__dict__ if isinstance(value, Waveform) else value) for key, value in waveform.__dict__.items()
        }
        return f"{waveform.__class__.__name__} {hashes}"



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Results class."""
from copy import deepcopy
from dataclasses import dataclass, field

import numpy as np
import pandas as pd

from qililab.constants import EXPERIMENT, RESULTSDATAFRAME, RUNCARD
from qililab.utils import coordinate_decompose
from qililab.utils.dataframe_manipulation import concatenate_creating_new_name_index
from qililab.utils.factory import Factory
from qililab.utils.loop import Loop
from qililab.utils.util_loops import compute_ranges_from_loops, compute_shapes_from_loops

from .counts import Counts
from .qblox_results.qblox_result import QbloxResult
from .result import Result


@dataclass
class Results:
    """Results class."""

    software_average: int
    num_schedules: int
    loops: list[Loop] | None = None
    shape: list[int] = field(default_factory=list)
    results: list[Result] = field(default_factory=list)
    _computed_dataframe_indices: list[str] = field(init=False, default_factory=list)
    _data_dataframe_indices: set[str] = field(init=False, default_factory=set)

    def __post_init__(self):
        """Add num_schedules to shape."""
        if not self.shape:
            self.shape = compute_shapes_from_loops(loops=self.loops)
        if self.num_schedules > 1:
            self.shape.append(self.num_schedules)
        if self.software_average > 1:
            self.shape.append(self.software_average)
        if self.results and isinstance(self.results[0], dict):
            tmp_results = deepcopy(self.results)
            # Pop the result name (qblox, ...) from the dictionary and instantiate its corresponding Result class.
            self.results = [Factory.get(result.pop(RUNCARD.NAME))(**result) for result in tmp_results]
        if self.loops is not None and isinstance(self.loops[0], dict):
            self.loops = [Loop(**loop) for loop in self.loops]

    def add(self, result: Result):
        """Append an ExecutionResults object.

        Args:
            result (Result): Result object.
        """
        self.results.append(result)

    def probabilities(self) -> dict[str, float]:
        """Probabilities of being in the ground and excited state of all the nested Results classes.

        Returns:
            dict[str, float]: Dictionary containing the probabilities (value) of being measured in each state (key).
        """
        return self._counts_object().probabilities()

    def counts(self):
        """Returns the counts dictionary containing the number of measurements (counts) of each state.

        Returns:
            dict[str, int]: Dictionary containing the number of measurements (value) in each state (key).
        """
        return self._counts_object().as_dict()

    def _counts_object(self) -> Counts:
        """Returns a Counts object containing the number of measurements (counts) of each state.

        Returns:
            Counts: Counts object containing the number of measurements (counts) of each state.
        """
        if len(self.results) == 0:
            return Counts(n_qubits=0)
        n_qubits = self.results[0].counts_object().n_qubits
        all_counts = Counts(n_qubits=n_qubits)
        for result in self.results:
            all_counts += result.counts_object()
        return all_counts

    def to_dataframe(self) -> pd.DataFrame:
        """Returns a single dataframe containing the info for the dataframes of all results. In the process, it adds an
        index that specifies to which result belongs the data.

        Returns:
            pd.DataFrame: List of probabilities of each executed loop and sequence.
        """

        result_dataframes = [result.to_dataframe() for result in self.results]
        return concatenate_creating_new_name_index(dataframe_list=result_dataframes, new_index_name="result_index")

    def _build_empty_result_dataframe(self):
        """Builds an empty result dataframe, with the minimal number of columns and nans as values"""
        return pd.DataFrame(
            [[np.nan] * len(self._data_dataframe_indices)],
            columns=pd.Index(self._data_dataframe_indices).transpose(),
            index=[0],
        ).reset_index(drop=True)

    def _concatenate_acquisition_dataframes(self):
        """Concatenates the acquisitions from all the results"""
        self._data_dataframe_indices = set().union(
            *[result.data_dataframe_indices for result in self.results if result is not None]
        )
        result_acquisition_list = [
            result.acquisitions().reset_index(drop=True) if result is not None else self._build_empty_result_dataframe()
            for result in self.results
        ]
        return concatenate_creating_new_name_index(
            dataframe_list=result_acquisition_list, new_index_name=RESULTSDATAFRAME.RESULTS_INDEX
        )

    def _generate_new_acquisition_column_names(self):
        """Checks shape, num_sequence and software_average and returns with that the list of columns that should
        be added to the dataframe."""
        new_columns = [
            f"{RESULTSDATAFRAME.LOOP_INDEX}{i}" for i in range(len(compute_shapes_from_loops(loops=self.loops)))
        ]
        if self.num_schedules > 1:
            new_columns.append(RESULTSDATAFRAME.SEQUENCE_INDEX)
        if self.software_average > 1:
            new_columns.append(RESULTSDATAFRAME.SOFTWARE_AVG_INDEX)
        return new_columns

    def _add_meaningful_acquisition_indices(self, result_acquisition_dataframe: pd.DataFrame) -> pd.DataFrame:
        """Add to the dataframe columns that are relevant indices, computable from the `result_index`, as:
        `loop_index_n` (in case more than one loop is defined), `sequence_index`"""
        old_columns = result_acquisition_dataframe.columns
        self._computed_dataframe_indices = self._generate_new_acquisition_column_names()
        self._data_dataframe_indices = set().union(
            *[result.data_dataframe_indices for result in self.results if result is not None]
        )

        result_acquisition_dataframe[self._computed_dataframe_indices] = result_acquisition_dataframe.apply(
            lambda row: coordinate_decompose(
                new_dimension_shape=self.shape,
                original_size=len(self.results),
                original_idx=row[RESULTSDATAFRAME.RESULTS_INDEX],
            ),
            axis=1,
            result_type="expand",
        )
        return result_acquisition_dataframe.reindex(
            columns=[*self._computed_dataframe_indices, *old_columns], copy=True
        )

    def _process_acquisition_dataframe_if_needed(
        self, result_dataframe: pd.DataFrame, mean: bool = False
    ) -> pd.DataFrame:
        """Process the dataframe by applying software average if required"""

        if mean and self.software_average > 1:
            preserved_columns = [
                col
                for col in result_dataframe.columns.values
                if col
                not in self._data_dataframe_indices.union(
                    {RESULTSDATAFRAME.RESULTS_INDEX, RESULTSDATAFRAME.SOFTWARE_AVG_INDEX}
                )
            ]
            groups_to_average = result_dataframe.groupby(preserved_columns)
            averaged_df = groups_to_average.mean().reset_index()
            averaged_df[RESULTSDATAFRAME.RESULTS_INDEX] = groups_to_average.first().reset_index()[
                RESULTSDATAFRAME.RESULTS_INDEX
            ]
            averaged_df.drop(columns=RESULTSDATAFRAME.SOFTWARE_AVG_INDEX, inplace=True)
            result_dataframe = averaged_df

        return result_dataframe

    def acquisitions(self, mean: bool = False) -> pd.DataFrame:
        """QbloxResult acquisitions of all the nested Results classes.

        Returns:
            np.ndarray: Acquisition values.
        """

        if self.results is None or len(self.results) <= 0:
            return pd.DataFrame([])

        if not isinstance(self.results[0], QbloxResult):
            raise ValueError(f"{type(self.results[0]).__name__} class doesn't have an acquisitions method.")

        self._fill_missing_values()

        result_acquisition_df = self._concatenate_acquisition_dataframes()
        expanded_acquisition_df = self._add_meaningful_acquisition_indices(
            result_acquisition_dataframe=result_acquisition_df
        )
        return self._process_acquisition_dataframe_if_needed(result_dataframe=expanded_acquisition_df, mean=mean)

    def _fill_missing_values(self):
        """Fill with None the missing values."""
        self.results += [None] * int(np.prod(self.shape) - len(self.results))

    @property
    def ranges(self) -> np.ndarray:
        """Results 'ranges' property.

        Returns:
            list: Values of the loops.
        """
        if self.loops is None:
            raise ValueError("Results doesn't contain a loop.")

        ranges = compute_ranges_from_loops(loops=self.loops)
        return np.array(ranges, dtype=object).squeeze()

    def to_dict(self) -> dict:
        """
        Returns:
            dict: Dictionary containing all the class information.
        """
        return {
            EXPERIMENT.SOFTWARE_AVERAGE: self.software_average,
            EXPERIMENT.NUM_SCHEDULES: self.num_schedules,
            EXPERIMENT.SHAPE: [] if self.loops is None else compute_shapes_from_loops(loops=self.loops),
            EXPERIMENT.LOOPS: [loop.to_dict() for loop in self.loops] if self.loops is not None else None,
            EXPERIMENT.RESULTS: [result.to_dict() for result in self.results],
        }

    @classmethod
    def from_dict(cls, dictionary: dict):
        """Transforms a dictionary into a Results instance. Inverse of to_dict().
        Args:
            dictionary: dict representation of a Results instance
        Returns:
            Results: deserialized Results instance
        """
        return Results(**dictionary)
